using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using SimpleJSON;
using System.Linq;

namespace ClockwiseSilver
{
	public class Gaze : MVRScript
	{
		private JSONStorableFloat recheckTimeJSON, randomTimeJSON, variationTimeJSON, randomH, randomV, randomD, randomChanceJSON,
			targetRangeJSON, posVariationJSON, followSpeedMinJSON, followSpeedMaxJSON, yOffsetJSON, uprightJSON,
			forceRandomMinJSON, forceRandomMaxJSON;
		private JSONStorableBool isActive, disablePosState, targetChange, includeSelf, enableLTrigger, enableHead,
			enableCamera, enableB3, enableGen3;
		private FreeControllerV3 headControl, neckControl;
		private FreeControllerV3.RotationState wasHeadRot = FreeControllerV3.RotationState.On;
		private FreeControllerV3.PositionState wasHeadPos = FreeControllerV3.PositionState.On;
		private Transform headTransform, chestTransform, gazeTarget;
		private List<Transform> targetList = new List<Transform>();
		private Vector3 lastGazePos, gazePos, lastT, t, randomT, currentRandomT, lastRandomT, randomPos = Vector3.zero;
		private Quaternion lastRot = Quaternion.identity;
		private bool gazeEnabled, isRandom = false;
		private float rotFactor, rotProgress, followFactor, lastFollowFactor, randomFactor, randomProgress,
			variationFactor, variationProgress = 1f;
		private float yOffset, zOffset, lastZOffset = 0f;
		
		public override void Init()
		{
			try
			{
				isActive = new JSONStorableBool ("isActive", false);
				RegisterBool (isActive);
				var toggle = CreateToggle(isActive);
				toggle.label = "Active";
				
				JSONStorableAction ToggleAction = new JSONStorableAction("Toggle Gaze", () =>
                {
                    isActive.val = !isActive.val;
                });
                RegisterAction(ToggleAction);
				
				JSONStorableAction StartAction = new JSONStorableAction("Start Gaze", () =>
                {
                    if (!isActive.val) { isActive.val = true; }
                });
                RegisterAction(StartAction);
				
				JSONStorableAction StopAction = new JSONStorableAction("Stop Gaze", () =>
                {
                    if (isActive.val) { isActive.val = false; }
                });
                RegisterAction(StartAction);
				
				JSONStorableAction RetargetAction = new JSONStorableAction("Random Gaze", () =>
                {
                    if (isActive.val) { ForceRandom(); }
                });
                RegisterAction(RetargetAction);
				
				targetChange = SetupToggle("Allow Target Change", true, true);
				recheckTimeJSON = SetupSlider("Recheck Time", 1.7f, 0.1f, 5f, true);
				randomTimeJSON = SetupSlider("Retarget Time", 5.5f, 1f, 20f, true);
				variationTimeJSON = SetupSlider("Variation Time", 4.4f, 1f, 20f, true);
				posVariationJSON = SetupSlider("Variation", 1f, 0f, 2f, true);
				
				followSpeedMinJSON = SetupSlider("Follow Speed Min", 0.45f, 0.1f, 0.6f, false);
				followSpeedMaxJSON = SetupSlider("Follow Speed Max", 0.85f, 0.625f, 1f, false);
				uprightJSON = SetupSlider("Stay Upright", 0.03f, 0, 0.06f, false);
				yOffsetJSON = SetupSlider("Y Offset", -0.1f, -1f, 1f, false);
				targetRangeJSON = SetupSlider("Target Range", 0f, -0.5f, 0.5f, false);

				CreateTextField(new JSONStorableString("header1", "Select Available Targets"),true).height = 10f;
				includeSelf = new JSONStorableBool ("Include Self", false); RegisterBool (includeSelf);
				toggle = CreateToggle(includeSelf, true);
				includeSelf.toggle.onValueChanged.AddListener((checkedVal) => { CreateTargetList(); });
				
				enableHead = new JSONStorableBool ("Target Head", true); RegisterBool (enableHead);
				toggle = CreateToggle(enableHead, true);
				enableHead.toggle.onValueChanged.AddListener((checkedVal) => { CreateTargetList(); });
				
				enableCamera = new JSONStorableBool ("Target Camera", true); RegisterBool (enableCamera);
				toggle = CreateToggle(enableCamera, true);
				enableCamera.toggle.onValueChanged.AddListener((checkedVal) => { CreateTargetList(); });
				
				enableLTrigger = new JSONStorableBool ("Target Female Gens", true); RegisterBool (enableLTrigger);
				toggle = CreateToggle(enableLTrigger, true);
				enableLTrigger.toggle.onValueChanged.AddListener((checkedVal) => { CreateTargetList(); });
				
				enableGen3 = new JSONStorableBool ("Target Male Gens", true); RegisterBool (enableGen3);
				toggle = CreateToggle(enableGen3, true);
				enableGen3.toggle.onValueChanged.AddListener((checkedVal) => { CreateTargetList(); });
				
				enableB3 = new JSONStorableBool ("Target Dildo", true); RegisterBool (enableB3);
				toggle = CreateToggle(enableB3, true);
				enableB3.toggle.onValueChanged.AddListener((checkedVal) => { CreateTargetList(); });
				
				var btn = CreateButton("Random Gaze", true);
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;
                btn.button.onClick.AddListener(() => { if (isActive.val) { ForceRandom(); } });
				
				forceRandomMinJSON = SetupSlider("Random Gaze Time Min", 0.35f, 0.1f, 2f, true);
				forceRandomMaxJSON = SetupSlider("Random Gaze Time Max", 0.8f, 0.1f, 2f, true);
				
				disablePosState = SetupToggle("Disable Head Pos Ctrl", false, false);
				
				CreateTextField(new JSONStorableString("header0", "Random Position Target"),false).height = 10f;
				randomChanceJSON = SetupSlider("Random Target Chance", 0.2f, 0f, 1f, false);
				randomH = SetupSlider("Random Target Horizontal", 0.8f, 0.1f, 4f, false);
				randomV = SetupSlider("Random Target Vertical", 0.4f, 0.1f, 4f, false);
				randomD = SetupSlider("Random Target Depth", 2f, 0.8f, 5f, false);
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}
		}
		
		private void Start()
		{
			StartCoroutine("StartRotuine");
		}
		
		private IEnumerator StartRotuine()
		{
			while (SuperController.singleton.isLoading)
			{
				yield return null;
			}
			
			gazeTarget = CameraTarget.centerTarget.transform;
			CreateTargetList();
			yield return null;
			
			if (isActive.val && !gazeEnabled)
			{
				bool allowStart = true;
				JSONStorable bjPlugin = FindPlugin("ClockwiseSilver.BJ");
				if (bjPlugin != null)
				{
					JSONStorableBool bjActive = bjPlugin.GetBoolJSONParam("isActive");
					if (bjActive != null && bjActive.val) { allowStart = false; }
				}
				
				if (allowStart) { StartGaze(); }
			}
			
			yield return null;
			
			//Set Listeners
			isActive.toggle.onValueChanged.AddListener(checkVal =>
				{
					if (checkVal) { StartGaze(); }
					else { StopGaze(); }
				});
			
			yOffset = yOffsetJSON.val;
			yOffsetJSON.slider.onValueChanged.AddListener((value) => { yOffset = yOffsetJSON.val; });
			variationTimeJSON.slider.onValueChanged.AddListener((value) =>
				{ if (value > randomTimeJSON.val) {randomTimeJSON.val = value;}; });
			randomTimeJSON.slider.onValueChanged.AddListener((value) =>
				{ if (value < variationTimeJSON.val) {variationTimeJSON.val = value;}; });
		}
		
		public void CreateTargetList()
		{
			bool wasActive = false;
			if (isActive.val) { wasActive = true; isActive.val = false; }
			targetList.Clear();
			if (enableCamera.val)
			{
				targetList.Add(CameraTarget.centerTarget.transform);
			}
			
			List<Atom> peopleAtoms = SuperController.singleton.GetAtoms().Where(a => a.category == "People").ToList();
			foreach (Atom a in peopleAtoms)
			{
				if (a == containingAtom && !includeSelf.val) { continue; }
				string aName = a.GetComponentInChildren<DAZCharacter>().name;
				
				foreach (Rigidbody rbd in a.rigidbodies)
				{
					if (rbd.name == "LabiaTrigger" && enableLTrigger.val && !aName.StartsWith("male"))
					{
						targetList.Add(rbd.transform);
					}
					if (rbd.name == "head" && enableHead.val && a != containingAtom)
					{
						targetList.Add(rbd.transform);
					}
					if ((rbd.name.ToLower() == "b3") && enableB3.val)
					{
						targetList.Add(rbd.transform);
					}
					if (rbd.name == "Gen3" && enableGen3.val && aName.StartsWith("male"))
					{
						targetList.Add(rbd.transform);
					}
				}					
			}
			
			isActive.val = wasActive;
		}
		
		private Vector3 GetRandomPos()
		{
			float rD = UnityEngine.Random.Range(0.75f, randomD.val);
			float rH = UnityEngine.Random.Range(-randomH.val, randomH.val);
			float rV = UnityEngine.Random.Range(-randomV.val, randomV.val);
			
			return headTransform.position
				+ (chestTransform.forward * rD)
				+ (chestTransform.right * rH)
				+ (chestTransform.up * rV);
		}
		
		private void SelectTarget()
		{
			if (!targetChange.val)
			{
				if (gazeTarget == null) { gazeTarget = CameraTarget.centerTarget.transform; }
				return;
			}
			
			if (UnityEngine.Random.Range(0f, 1f) < randomChanceJSON.val)
			{
				isRandom = true;
				randomPos = GetRandomPos();
				return;
			}
			isRandom = false;
			
			if (targetList.Count == 0) { gazeTarget = CameraTarget.centerTarget.transform; return; }
			
			int targetIndex = (int)UnityEngine.Random.Range(0, targetList.Count);
			if (targetIndex > targetList.Count - 1) { targetIndex = targetList.Count - 1; }
			
			Vector3 headPos = headTransform.position;
			Vector3 refForward = chestTransform.forward;
			Transform testTarget = targetList[targetIndex];

			if (Vector3.Dot(testTarget.position - headPos, refForward) < targetRangeJSON.val)
			{
				bool found = false;
				for (int i = 0; i < targetList.Count; i++)
				{
					if (i == targetIndex) { continue; }
					testTarget = targetList[i];
					if (Vector3.Dot(testTarget.position - headPos, refForward) >= targetRangeJSON.val)
					{
						gazeTarget = testTarget;
						found = true;
						return;
					}
				}
				if (!found)
				{
					isRandom = true;
					randomPos = GetRandomPos();
					return;
				}
			}
			
			gazeTarget = testTarget;
		}
		
		private void ForceRandom()
		{
			isRandom = true;
			randomPos = GetRandomPos();
			lastT = t;
			if (forceRandomMinJSON.val > forceRandomMaxJSON.val)
			{
				forceRandomMinJSON.val = forceRandomMaxJSON.val;
			}
			float recheckTimer = UnityEngine.Random.Range(forceRandomMinJSON.val, forceRandomMaxJSON.val);
			rotFactor = 1f / recheckTimer;
			rotProgress = 0f;
		}
		
		private void SetTarget()
		{
			UnityEngine.Random.InitState((int)Time.time);
			
			SelectTarget();
			lastRandomT = currentRandomT;
			randomT.x = UnityEngine.Random.Range(-7.5f, 7.5f) * posVariationJSON.val;
			randomT.y = UnityEngine.Random.Range(-7.5f, 7.5f) * posVariationJSON.val;
			randomT.z = UnityEngine.Random.Range(-7.5f, 7.5f) * posVariationJSON.val;
			
			lastZOffset = zOffset;
			Vector3 d = Vector3.Normalize(gazeTarget.position - headTransform.position);
			zOffset = -Vector3.SignedAngle(headTransform.up, Vector3.up, d) * uprightJSON.val;
			zOffset = Mathf.Clamp(zOffset, -1.5f, 1.5f);
			
			lastFollowFactor = followFactor;
			followFactor = UnityEngine.Random.Range(followSpeedMinJSON.val, followSpeedMaxJSON.val);
			
			float randomTimer = randomTimeJSON.val * UnityEngine.Random.Range(0.8f, 1.2f);
			randomFactor = 1f / randomTimer;
			randomProgress = 0f;
			
			float variationTimer = variationTimeJSON.val * UnityEngine.Random.Range(0.8f, 1.2f);
			if (variationTimer > randomTimer) { variationTimer = randomTimer; }
			variationFactor = 1f / variationTimer;
			variationProgress = 0f;
		}
		
		private void Recheck()
		{
			lastT = t;
			lastGazePos = gazePos;
			gazePos = gazeTarget.position;
			float recheckTimer = recheckTimeJSON.val * UnityEngine.Random.Range(0.8f, 1.2f);
			rotFactor = 1f / recheckTimer;
			rotProgress = 0f;
		}
		
		private void StartGaze()
		{
			if (gazeEnabled) { return; }
			
			SyncAtom();
			
			wasHeadPos = headControl.currentPositionState;
			wasHeadRot = headControl.currentRotationState;
			headControl.currentRotationState = FreeControllerV3.RotationState.Off;
			
			if (disablePosState.val)
			{
				headControl.currentPositionState = FreeControllerV3.PositionState.Off;
			}
			
			neckControl.currentRotationState = FreeControllerV3.RotationState.Off;
			neckControl.currentPositionState = FreeControllerV3.PositionState.Off;
			neckControl.jointRotationDriveSpring = 75f;
			neckControl.jointRotationDriveMaxForce = 75f;
			
			SetTarget();
			t = gazeTarget.position;
			Recheck();
			
			gazeEnabled = true;
		}
		
		private void StopGaze()
		{
			if (!gazeEnabled) { return; }
			gazeEnabled = false;
			
			headControl.jointRotationDriveXTarget = 0f;
			headControl.jointRotationDriveYTarget = 0f;
			headControl.jointRotationDriveZTarget = 0f;
			neckControl.jointRotationDriveXTarget = 0f;
			neckControl.jointRotationDriveYTarget = 0f;
			neckControl.jointRotationDriveZTarget = 0f;
			neckControl.jointRotationDriveSpring = 20f;
			neckControl.jointRotationDriveMaxForce = 25f;
			
			headControl.currentRotationState = wasHeadRot;
			if (disablePosState.val)
			{
				headControl.currentPositionState = wasHeadPos;				
			}
		}
		
		private void Update()
		{
			if (!gazeEnabled) { return; }
			float dt = Time.deltaTime;
			
			rotProgress = Mathf.Clamp01(rotProgress + dt * rotFactor);
			randomProgress = Mathf.Clamp01(randomProgress + dt * randomFactor);
			if (rotProgress >= 1f)
			{
				if (randomProgress >= 1f) { SetTarget(); }
				Recheck();
			}
			
			if (isRandom)
			{
				t = randomPos;
				t = Vector3.Lerp(lastT, t, QuadraticInOut(rotProgress));
			}
			else
			{
				float ff = Mathf.Lerp(lastFollowFactor, followFactor, QuadraticInOut(randomProgress));
				t = Vector3.Lerp(gazePos, gazeTarget.position + (chestTransform.up * yOffset), ff);
				t = Vector3.Lerp(lastT, t, QuadraticInOut(rotProgress));			
			}
			
			Vector3 d = Vector3.Normalize(t - headTransform.position);
			d = chestTransform.InverseTransformDirection(d);
			
			variationProgress += dt * variationFactor;
			float vp = QuartInOut(Mathf.Clamp01(variationProgress));
			currentRandomT.x = Mathf.Lerp(lastRandomT.x, randomT.x, vp);
			currentRandomT.y = Mathf.Lerp(lastRandomT.y, randomT.y, vp);
			currentRandomT.z = Mathf.Lerp(lastRandomT.z, randomT.z, vp);
			
			float zo = Mathf.Lerp(lastZOffset, zOffset, QuartInOut(randomProgress));
			
			headControl.jointRotationDriveXTarget = currentRandomT.x + (d.y * 30f);
			headControl.jointRotationDriveYTarget = currentRandomT.y + (d.x * -25f);
			headControl.jointRotationDriveZTarget = currentRandomT.z + (d.x * 7.5f) + (zo * 7.5f);
			neckControl.jointRotationDriveXTarget = currentRandomT.x + (d.y * 35f);
			neckControl.jointRotationDriveYTarget = currentRandomT.y + (d.x * -30f);
			neckControl.jointRotationDriveZTarget = currentRandomT.z + (d.x * 10f) + (zo * 10f);
		}
		
		private void SyncAtom()
		{
			headControl = containingAtom.freeControllers.First(fc => fc.name == "headControl");
			headTransform = headControl.transform;
			neckControl = containingAtom.freeControllers.First(fc => fc.name == "neckControl");
			chestTransform = containingAtom.rigidbodies.First(rb => rb.name == "chest").transform;
		}
		
		private JSONStorableFloat SetupSlider(string name, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(name, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateSlider(storable, rightSide);
			RegisterFloat(storable);
			return storable;
		}
		
		private JSONStorableBool SetupToggle(string name, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(name, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateToggle(storable, rightSide);
			RegisterBool(storable);
			return storable;
		}
		
		private float QuartInOut(float x)
		{
			if (x < 0.5f) { return 8 * x * x * x * x; }
			return 1 - Mathf.Pow(-2 * x + 2, 4) / 2;
		}
		
		private float QuadraticInOut(float x)
		{
			if ((x *= 2f) < 1f) { return 0.5f * x * x; }
            return -0.5f * ((x -= 1f) * (x - 2f) - 1f);
		}
		
		private JSONStorable FindPlugin(string fullClassName)
		{
			string pluginName = containingAtom.GetStorableIDs().Find(s => s.StartsWith("plugin#") && s.Contains(fullClassName));
			if (pluginName != null && pluginName != "")
			{
				return containingAtom.GetStorableByID(pluginName);
			}
			else
			{
				return null;
			}
		}
		
		private void OnDisable()
		{
			StopGaze();
		}
	}
}